DEFAULT_BRANCH = "main"
# Single merge to main
TEST_01_MIGRATION_BRANCH = "feature/migration-test-01"
# Single merge to main after 01 is merged (single conflict)
TEST_02_MIGRATION_BRANCH = "feature/migration-test-02"
# Single merge to main after 01 is merged (multiple conflicts)
TEST_03_MIGRATION_BRANCH = "feature/migration-test-03"
# Single merge to main after 01 is merged (named module)
TEST_04_MIGRATION_BRANCH = "feature/migration-test-04"
# Multiple migrations with a named migration file
TEST_05_MIGRATION_BRANCH = "feature/migration-test-05"
